from . import DataPrepare
from . import OSKit
from . import Stats
from . import WaybackScraper
from . import CrawlerApi
from . import _BasicFunc
from ._version import __version__

"""
package principle:
    1. Input and output should end with input/return type
       Like list/arr/df/bool_arr/dict/int/float/str...
    2. constant vars should be UPPER_CASE, others be snake_case
    3. IMPORTANT: ALL CHANGEABLE TYPE(df,array,list) should 
        DEEP COPY before CHANGED IN MEMORY!
"""



import time
import functools


def timer_wrapper(func):
    """Wrapper for function running timing"""
    @functools.wraps(func)
    def decorated(*args, **kwargs):
        time_start = time.time()
        result = func(*args, **kwargs)
        time_end = time.time()
        time_spend = time_end - time_start
        print('%s cost time: %.5f s' % (func.__name__, time_spend))
        return result
    return decorated



import numpy as np
import pandas as pd


def test_allsame_bool(listarrser1, listarrser2):
    """
    test all be same
    :return: bool, is all same or not all same
    """
    temp_df = pd.DataFrame({'ser1': listarrser1, 'ser2': listarrser2})
    temp_df = temp_df.dropna(axis=0)
    temp_len = temp_df.shape[0]

    return np.nansum(temp_df.apply(func=lambda x: x['ser1'] == x['ser2'], axis=1).values) == temp_len


def repunit_append_arr(target_len, listarrser, repunit):
    """
    Append the repeat units to the end if the listarr to the target length
    :param target_len: target length of the list/array
    :param listarrser: input should be list/arr/ser
    :return: arr
    """
    temp_arr = np.array(listarrser) # even series change to array
    sup_arr = np.repeat(repunit, target_len - len(temp_arr))
    return np.concatenate([temp_arr, sup_arr], axis=0)



import tqdm
import numpy as np
import pandas as pd
import pathos
import difflib
import rapidfuzz
import time
import thefuzz.process
from .. import _BasicFunc


def list_fuzzymatching_df(querying_listarr, choice_list, method, scorer):
    """
    lisklikearray->fuzzymatch with a LIST. scorer I prefer rapidfuzz.fuzz.ratio(Normalize levenshtein)
    :param querying_listarr:~
    :param choice_list:~
    :param method: npapply/multiprocessing/rapidfuzz_cdist/difflib,difflib use special method to match, Not levenshtein
    :param scorer: the scorer of fuzzymatch, it could be changed to different result,
            classic normalized_levenshtein be rapidfuzz.fuzz.ratio,however be useless in difflib method
    :return: result dataframe
    """

    # 0.1.1.230205_alpha-> new version only check the n.a. of query list, automatically remove choices duplicates/na >>>
    """
    deprecated old version, 0.1.1.230203_alpha
    
    if (sum(pd.Series(querying_listarr).isna()) != 0) | (sum(pd.Series(choice_list).isna()) != 0):
        raise ValueError(
            "Error:querying_listarr/choice_list are not ZERO NULL/NAN OBS:'\n" +
            f"querying_listarr {sum(pd.Series(querying_listarr).isna())} NA," +
            f" choice_list {sum(pd.Series(choice_list).isna())} NA"
        )
    
    if np.any(pd.Series(choice_list).duplicated()):
        raise ValueError('Error:there are duplicated in choice list, may cause confusing result')
    """

    if np.any(pd.Series(querying_listarr).isna()):
        raise ValueError("\033[31mNull/Na in querying list may cause confounding\033[0m")

    if np.any(pd.Series(choice_list).duplicated()) or np.any(pd.Series(choice_list).isna()):
        print(
            "\nDuplicates or Null/Na in choices are detected and automatically removed\n"
        )
        choice_list = pd.Series(choice_list).dropna().drop_duplicates().to_list()
    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

    # set constants
    """CONSTANTS CAN NOT BE CHANGED"""
    CONST_querying_listarr = pd.Series(querying_listarr).astype(str).tolist()
    CONST_choice_list = pd.Series(choice_list).astype(str).tolist()

    def __LAMBDA_taskof_list_fuzzymatching_list_tuple(identifier):
        """
        :param identifier:identifier
        :return: results tuple
        """
        value = thefuzz.process.extractOne(
            query=CONST_querying_listarr[identifier],
            choices=CONST_choice_list,
            scorer=scorer
        )
        return (value, identifier)

    """Dropped, for even slower then npapply and totally unstable, WRONG OUTPUTS MOSTTIMES"""
    # def __LAMBDA_taskof_list_rapidFM_list_tuple(identifier):
    #     """
    #     :param identifier:identifier
    #     :return: results tuple
    #     """
    #     value = rapidfuzz.process.extractOne(query=CONST_querying_listarr[identifier],
    #                                         #Only Choose first two rows
    #                                          choices=CONST_choice_list,
    #                                          scorer=scorer
    #                                          )[:2]
    #     return (value,identifier)

    def __LAMBDA_taskof_list_difflibmatching_list_tuple(identifier):
        """
        :param identifier:
        :return:
        """
        temp_input = CONST_querying_listarr[identifier]
        match_result = difflib.get_close_matches(word=temp_input, possibilities=CONST_choice_list, n=1)
        if match_result == []:
            match_result = ''
        else:
            match_result = match_result[0]
        match_scores = difflib.SequenceMatcher(None, match_result, temp_input).ratio()
        #  *100
        value = (match_result, match_scores)

        return (value, identifier)

    if method == 'npapply':
        npapply_query_arr = np.array(CONST_querying_listarr)[:, np.newaxis]

        time_start = time.time()
        npapply_match_result_mat = np.apply_along_axis(
            func1d=lambda x: thefuzz.process.extractOne(query=x[0], choices=CONST_choice_list, scorer=scorer),
            arr=npapply_query_arr, axis=1)
        time_end = time.time()
        print('list_fuzzymatching_list method {} time cost'.format(method), time_end - time_start, 's')

        #  In after place match mean result from choice list
        #  Only Choose first two rows
        match_result_df = pd.DataFrame(npapply_match_result_mat, columns=['match_list', 'match_score'])

    elif method in ['multiprocessing', 'difflib']:
        #  Slow when compare with others, However is useful
        _temp_loop_function = {
            'multiprocessing': __LAMBDA_taskof_list_fuzzymatching_list_tuple,
            # 'rapidfuzz_multiprocessing': __LAMBDA_taskof_list_rapidFM_list_tuple,
            'difflib': __LAMBDA_taskof_list_difflibmatching_list_tuple
        }[method]

        temp_matching_list = []
        temp_score_list = []
        temp_identifier_list = []

        with pathos.multiprocessing.Pool(
                # for safer exception in multiprocess
                initializer=_BasicFunc.MultiprocessF.threads_interrupt_initiator
        ) as pool:
            for result in tqdm.tqdm(
                    pool.imap_unordered(
                        _temp_loop_function,
                        range(len(CONST_querying_listarr))
                    ),
                    total=len(CONST_querying_listarr)
            ):
                temp_matching_list.append(result[0][0])
                temp_score_list.append(result[0][1])
                temp_identifier_list.append(result[1])

        #  In after place match mean result from choice list
        mp_match_result_df = pd.DataFrame(
            {
                'match_list': temp_matching_list,
                'match_score': temp_score_list,
                'identifier': temp_identifier_list

            }
        )
        mp_match_result_df = mp_match_result_df.sort_values(by='identifier', ascending=True)
        match_result_df = mp_match_result_df[['match_list', 'match_score']].reset_index().copy()
        del match_result_df['index']

    elif method == 'rapidfuzz_cdist':
        """SuperFast Cpp Matrix method for matching result"""
        time_start = time.time()
        matchscores_total_mat = rapidfuzz.process.cdist(queries=CONST_querying_listarr,
                                                        scorer=scorer,
                                                        choices=CONST_choice_list,
                                                        workers=-1)
        matchscores_position_arr = np.nanargmax(matchscores_total_mat, axis=1)
        # In after place match mean result from choice list
        match_result_df = pd.DataFrame(
            {
                'match_list': np.array(CONST_choice_list)[matchscores_position_arr],
                'match_score': matchscores_total_mat[np.arange(len(CONST_querying_listarr)),
                                                     matchscores_position_arr],
            }
        )
        time_end = time.time()
        print('list_fuzzymatching_list method {} time cost'.format(method), time_end - time_start, 's')
    else:
        raise ValueError('Wrong method of iteration')

    match_result_df = match_result_df.copy()
    match_result_df['match_score'] = match_result_df['match_score'].astype(float)
    # make 0~100%
    if np.all((match_result_df['match_score'].values >= 0) &
              (match_result_df['match_score'].values <= 1)):
        match_result_df['match_score'] = np.multiply(match_result_df['match_score'], 100)
        print(f'method {method},scorer {scorer} has *100 their scores')

    return match_result_df


# ----------------------------------------------------------------------------------------------------------------------
# """L1 Complex DataPrepare Functions"""################################################################################
# ----------------------------------------------------------------------------------------------------------------------


def l1_auto_fuzzymatching_df(querying_listarr, choice_list, slicing_len, method, scorer):
    """
    slicing for memory saving....
    lisklikearray->fuzzymatch with a LIST.scorer I prefer rapidfuzz.fuzz.ratio(Normalize levenshtein)
    :param slicing_len: slicing len for each matching(rows)
    :param querying_listarr:~
    :param choice_list:~
    :param method: npapply/multiprocessing/rapidfuzz_cdist/difflib
                ,difflib use special method to match, Not levenshtein
    :param scorer: the scorer of fuzzymatch, it could be changed to different result,
            classic normalized_levenshtein be fuzz.ratio,however be useless in difflib method
    :return: result dataframe
    """

    # 0.1.1.230205_alpha-> new version only check the n.a. of query list, automatically remove choices duplicates/na >>>

    # add this only for not raise Value-error after running long-time script
    if np.any(pd.Series(querying_listarr).isna()):
        raise ValueError("\033[31mAuto Function->Null/Na in querying list may cause confounding\033[0m")

    # add this to avoid waste of performance and consistency
    if np.any(pd.Series(choice_list).duplicated()) or np.any(pd.Series(choice_list).isna()):
        print(
            "\nDuplicates or Null/Na in choices are detected and automatically removed\n"
        )
        choice_list = pd.Series(choice_list).dropna().drop_duplicates().to_list()
    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    querying_arr = np.array(querying_listarr)
    # slicing for memory easier use.
    slice_arr = np.arange(0, len(querying_arr), slicing_len)
    print(
        'TOTAL LENGTH of none SLICED data:{},SLICES be {} seps'.format(len(querying_arr), len(slice_arr))
    )
    if slice_arr[-1] != len(querying_arr):
        slice_arr = np.append(slice_arr, len(querying_arr))   # his step is run in 100% but not remove 'if'

    df_concat_list = []
    for k in range(len(slice_arr) - 1):
        print('MATCH SLICE:{}~{}'.format(slice_arr[k], slice_arr[k + 1]))
        querying_k_arr = querying_arr[slice_arr[k]:slice_arr[k + 1]]
        matchresult_k_df = list_fuzzymatching_df(querying_k_arr, choice_list, method=method, scorer=scorer)

        df_concat_list.append(matchresult_k_df)

    total_df = pd.concat(df_concat_list, axis=0)
    return total_df



import signal


def threads_interrupt_initiator():
    """
    Function for multiprocessing.Pool(initializer=threads_interrupt_initiator())
    Each pool process will execute this as part of its initialization.
    Use this to keep safe for multiprocessing...and gracefully interrupt by keyboard
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)



"""
Refactor from https://github.com/r-boulland/Corporate-Website-Disclosure.git
Manually refactor:

Author: Qihang Zhang in 2023/02/03,National University of Singapore,
NUS Business School, Department of Accounting
"""
from . import WaybkJsonToCsvT
from . import cls_WayBack



import numpy as np
import pandas as pd
import warnings


def summary_quantile_ser(arrser):
    try:
        return pd.Series({
            'Mean': np.nanmean(arrser),
            'Sd': np.nanstd(arrser),
            'p25': np.nanquantile(arrser, q=0.25),
            'Median': np.nanmedian(arrser),
            'p75': np.nanquantile(arrser, q=0.75)
        })
    except:
        warnings.warn('Some Error in selfmade, use pd.describe instead',
                      SyntaxWarning)
        return pd.Series(arrser).describe()




import json
import csv
import pandas as pd
import warnings
import os
"""
modified from origin WaybkJsonToCsvT.py

Refactor: Qihang Zhang in 2023/02/03,National University of Singapore,
NUS Business School, Department of Accounting

"""


def wayback_json_to_csv(input_json_path: str, output_csv_path: str):
    if not pd.Series(input_json_path).str.endswith("json")[0]:
        raise ValueError('Input must endswith .json')
    # read json file
    data_file = open(input_json_path)
    data = json.load(data_file)
    data_file.close()

    # start csv
    csv_file = open(output_csv_path, 'w', newline='')
    # print(output_csv_path)
    writer = csv.writer(csv_file)

    write_counter = 0
    for element in data:
        try:
            writer.writerow(element)
        except:
            warnings.warn(f"csv.writer.writerow error in {write_counter}'s element(minimum zero)")
        finally:
            write_counter += 1

    csv_file.close()


def dir_wayback_json_to_csv(input_dir, output_dir):
    """see more in https://github.com/r-boulland/Corporate-Website-Disclosure.git"""
    for filename in os.listdir(input_dir):
        if not pd.Series(filename).str.endswith(".json")[0]:
            continue
        out = os.path.join(output_dir, os.path.splitext(filename)[0])
        seq = (out, ".csv")
        out = "".join(seq)
        ticker = os.path.join(input_dir, filename)
        print('reading:', ticker)
        data_file = open(ticker)
        try:
            data = json.load(data_file)
        except:
            continue
        data_file.close()
        csv_file = open(out, 'w', newline='')
        print('writing:', out, '\n')
        writer = csv.writer(csv_file)
        for element in data:
            try:
                writer.writerow(element)
            except:
                continue
        csv_file.close()



import os
import numpy as np
import pandas as pd
import warnings
from .. import _BasicFunc


def check_make_directory(check_make_dir, force: bool = False):
    """
    Check and Make directory
    :param check_make_dir: the directory you hope to check-make
    :param force: is force to make the recursive dirs
    """
    folder = os.path.exists(check_make_dir)

    if not folder:
        if force:
            warnings.warn("\033[31myou already set 'force=True' for recursive, however unsafe\033[0m", UserWarning)
            temp_folder_maker = os.makedirs
        else:
            temp_folder_maker = os.mkdir

        temp_folder_maker(check_make_dir)

        print('Make dir:{}'.format(check_make_dir))
    else:
        print('{} already exist.'.format(check_make_dir))


def dir_colnames_df(read_df_func, read_dir, suffix_regex: str, **kwargs):
    """
    To show the colnames of Each DF in a dir, summary in to an DF

    :param read_df_func: the function to read the datas in the dir, function(dir+names),
        equally function(path), however,read_function must return pd.DataFrame...
    :param read_dir: the directory of the datas saved
    :param suffix_regex: suffix regex to search for files, however, should be type in "regex" to search.
        if you need to set Nothing, then simply use '', None, etc...
    :param kwargs: other arguments for suffix regex detect file names which pandas.Series.str.contains() use, see:
        https://pandas.pydata.org/docs/reference/api/pandas.Series.str.contains.html,
        however, regex=True,pat=patterns(special) can not be modified.
    :return: dataframe of output cols
    """
    colname_dict = {}
    colname_len_dict = {}

    file_arr = np.array(os.listdir(read_dir))

    # if we are sure that the suffix regex is string, then set it to suffix check.
    try:
        if (len(suffix_regex) != 0) and (isinstance(suffix_regex, str)):
            pattern = '^.*(' + suffix_regex + '){1}$'
            temp_choose_arr = pd.Series(file_arr).str.contains(
                pat=pattern, regex=True, **kwargs
            ).values
            file_arr = file_arr[temp_choose_arr]
        else:
            print('\033[31mfail in suffix regex using, \nhowever pass file suffix detecting, use all files\033[0m')
    except Exception as e:
        print(f'\033[31m{e} occurs in suffix regex using, \nhowever pass file suffix detecting, use all files\033[0m')

    # print the detect result of the function...
    print('FILENAMES:\n', file_arr)

    # get read_sample to ensure the readfunc return dataframe
    read_sample = read_df_func(read_dir + file_arr[0])
    if not isinstance(read_sample, pd.DataFrame):
        raise ValueError('read_function must return pandas.Dataframe, '
                         f'However, your read_function return {type(read_sample)}'
                         'reset read function!')

    for i in file_arr:
        # read each
        colname_dict[i] = read_df_func(read_dir + i).columns.values
        colname_len_dict[i] = len(colname_dict[i])
    for i in file_arr:
        colname_dict[i] = _BasicFunc.ArrayLikeF.repunit_append_arr(
            target_len=max(colname_len_dict.values()),
            listarrser=colname_dict[i],
            repunit=np.nan)

    temp_df = pd.DataFrame(colname_dict)

    return temp_df



import pandas as pd
import requests


def api_usps_crosswalk_df(digital_token: str,
                          type_id: int,
                          query_id,
                          year: int,
                          quarter: int
                          ):
    """
    More information see https://www.huduser.gov/portal/dataset/uspszip-api.html

    :param digital_token: register to upper url for special token
    :param type_id:
    :param query_id:
    :param year:
    :param quarter:1,2,3,4
    :return: dataframe
    """
    if not (quarter in [1, 2, 3, 4]):
        raise ValueError('quarter should in 1,2,3,4; and it should be int')
    response_map_dict = {
        400: 'An invalid value was specified for one of the query parameters in the request URI.',
        401: 'Authentication failure',
        403: 'Not allowed to access this dataset API because you have not registered for it.',
        404: "No data found using '(value you entered)'",
        405: "Unsupported method, only GET is supported",
        406: "Unsupported Accept Header value, must be application/json",
        500: "Internal server error occurred"
    }

    url = f"https://www.huduser.gov/hudapi/public/usps?type={type_id}&query={query_id}&year={year}&quarter={quarter}"
    token = digital_token
    headers = {"Authorization": "Bearer {0}".format(token)}

    response = requests.get(url, headers=headers)

    if response.status_code != 200:
        raise ValueError(f"{response.status_code} Failure:{response_map_dict[response.status_code]}")

    else:
        temp_df = pd.DataFrame(response.json()["data"]["results"])

    return temp_df



import numpy as np
import pandas as pd


def sic4_fininst_select_bool_arr(listarrser):
    #2022/12/15
    # temp_arr = pd.Series(listarrser).astype('int32').values
    #Set again:
    # pandas.errors.IntCastingNaNError: Cannot convert non-finite values (NA or inf) to integer
    temp_ser = pd.Series(listarrser).copy()
    bool_notNA_ser = ~(temp_ser.isna())
    temp_ser[bool_notNA_ser] = (temp_ser[bool_notNA_ser]).apply('int32')

    bool_arr = (np.greater_equal(temp_ser.values,6000) & np.less_equal(temp_ser.values,6999))
    return bool_arr


def naic6_fininst_select_bool_arr(listarrser):
    #Set again:
    # pandas.errors.IntCastingNaNError: Cannot convert non-finite values (NA or inf) to integer
    temp_ser = pd.Series(listarrser).copy()
    bool_notNA_ser = ~(temp_ser.isna())
    temp_ser[bool_notNA_ser] = (temp_ser[bool_notNA_ser]).apply('int32')

    bool_arr = (np.greater_equal(temp_ser.values,520000) & np.less_equal(temp_ser.values,529999))
    return bool_arr


def int_rangeselect_bool_arr(listarrser, floor:int, ceil:int,
                             include_floor_bool:bool,
                             include_ceil_bool:bool
                             ) :
    """
    Universal function,select bool array from already exist listarrser, Likely to "NAIC_bank_select_bool_arr"

    :param listarrser: the query listarrser(arraylike)
    :param floor:the floor of the range selection, be int
    :param ceil: the ceil of the range selection, be int
    :param include_floor_bool: include floor or not
    :param include_ceil_bool: include ceil or not
    :return: bool arr for selection outcomes
    """
    if isinstance(include_floor_bool,bool) and isinstance(include_ceil_bool,bool):
        pass
    else:
        raise ValueError('include_floor_bool,include_ceil_bool should be bool')

    temp_floor_func = np.greater_equal if include_floor_bool else np.greater
    temp_ceil_func = np.less_equal if include_ceil_bool else np.less
    #Set again:
    # pandas.errors.IntCastingNaNError: Cannot convert non-finite values (NA or inf) to integer
    temp_ser = pd.Series(listarrser).copy()
    bool_notNA_ser = ~(temp_ser.isna())
    temp_ser[bool_notNA_ser] = (temp_ser[bool_notNA_ser]).apply('int32')

    bool_arr = (temp_floor_func(temp_ser.values,floor) & temp_ceil_func(temp_ser.values,ceil))
    return bool_arr

# -----------------------------------------------------------------------------------------
# """L1 Complex DataPrepare Functions"""#######################################################
# -----------------------------------------------------------------------------------------


def l1_fininst_sic4naic6_selectbool_arr(SIC_listarrser, NAIC_listarrser, combine_method):
    """
    :param SIC_listarrser: If None,then set None, or listarrser
    :param NAIC_listarrser: If None,then set None, or listarrser
    :param combine_method: how to combine the result if input SIC and NAIC together in the SAME INPUT."and',"or"
    :return: np.array
    """
    if (SIC_listarrser is None) and (NAIC_listarrser is None):
        raise ValueError('You have to input least SIC/NAIC')
    elif (not (SIC_listarrser is None)) and (not (NAIC_listarrser is None)):
        SIC_bool_arr = sic4_fininst_select_bool_arr(SIC_listarrser)
        NAIC_bool_arr = naic6_fininst_select_bool_arr(NAIC_listarrser)
        if combine_method == 'and':
            bool_arr = SIC_bool_arr & NAIC_bool_arr
        elif combine_method == 'or':
            bool_arr = SIC_bool_arr | NAIC_bool_arr
        else:
            raise ValueError('combine_method Invalid,input "and","or" instead')
    elif not (SIC_listarrser is None):
        bool_arr = sic4_fininst_select_bool_arr(SIC_listarrser)
    elif not (NAIC_listarrser is None):
        bool_arr = naic6_fininst_select_bool_arr(NAIC_listarrser)
    else:
        raise ValueError('Unexpected input')

    return bool_arr



import numpy as np
import pandas as pd
import copy
import typing


def na_percent_float(listarrser) -> float:
    temp_ser = pd.Series(listarrser)
    temp_len = len(temp_ser)
    na_percentage = sum(pd.isna(temp_ser)) / temp_len
    return na_percentage


def remove_duplicate_na_arr(listarrser):
    """remove the duplicate obs and NA obs"""
    temp_ser = pd.Series(listarrser).drop_duplicates().dropna()
    return temp_ser.values


def latent_na_detector_bool_arr(listarrser, na_regex, case, is_detect_explicit_na: bool):
    """
    detect latent na, like spaces or comma
    :param is_detect_explicit_na: bool, True for detect explicit na, False for not detect
    :param listarrser: input
    :param na_regex: regex
    :param case: If True, case sensitive.
    :return: bool array
    """
    if not isinstance(is_detect_explicit_na,bool):
        raise ValueError('is_detect_explicit_na only input for True or False')

    na_detect_arr = pd.Series(listarrser).str.match(
        pat='^((,)|(\.)|( ))*('+na_regex+'){1}((,)|(\.)|( ))*$',
        case=case,
        na=is_detect_explicit_na  # substitute explicit n.a. with True/False
    ).values

    return na_detect_arr


def auto_substituite_duplicates_bychoices_arr(duplicates_listarrser, choice_listarrser,
                                              duplicate_keep: typing.Literal['first', 'last', False]):
    """
    Use choice list in choice_listarrser to help the duplicates substitute the found duplicates;
    However, Auto looping for no substitute result.
    :param duplicates_listarrser: target
    :param choice_listarrser:choices, length must same as target listarrser
    :param duplicate_keep: the method of pd.duplicates() keep, the True ones should be substituted
    :return: np.array
    """

    duplicates_arr = np.array(duplicates_listarrser)
    choice_arr = np.array(choice_listarrser)

    if len(choice_arr) != len(duplicates_arr):  # length check
        raise ValueError('length of duplicates_listarrser and choice_listarrser are not same')

    if pd.Series(duplicates_arr).nunique() != len(duplicates_arr):
        print('duplicates_listarrser has no duplicates')  # check duplicates
        return duplicates_arr

    if not (duplicate_keep is None):  # loop
        # Initialize
        temp_duplicates_boolean_arr = pd.Series(duplicates_arr).duplicated(keep=duplicate_keep).values
        temp_nodup_remove_suffix_arr = copy.deepcopy(duplicates_arr)
        temp_nodup_remove_suffix_arr[temp_duplicates_boolean_arr] = choice_arr[temp_duplicates_boolean_arr]
        # other arguments
        _temp_loop_count = 0
        _temp_keep = duplicate_keep
        _temp_duplicate_count = sum(temp_duplicates_boolean_arr)
        print(f'Initialize autorestore duplicates last for:{sum(temp_duplicates_boolean_arr)},\n')
        # deal with duplicates
        while pd.Series(temp_nodup_remove_suffix_arr).nunique() != len(temp_nodup_remove_suffix_arr):

            last_loop_arr = temp_nodup_remove_suffix_arr
            temp_duplicates_boolean_arr = pd.Series(last_loop_arr
                                                    ).duplicated(keep=_temp_keep).values
            temp_nodup_remove_suffix_arr = copy.deepcopy(last_loop_arr)
            temp_nodup_remove_suffix_arr[temp_duplicates_boolean_arr] = choice_arr[temp_duplicates_boolean_arr]
            _temp_loop_count += 1
            # check if in INFINITE LOOP: IF same, means last time dealing is meaningless
            if _temp_duplicate_count == sum(temp_duplicates_boolean_arr):
                if _temp_keep != False:
                    print('To stop infinite loop, change to False')
                    _temp_keep = False
                else:
                    raise ValueError('Problems in choice list, It cause infinite loop unsolvable')

            _temp_duplicate_count = sum(temp_duplicates_boolean_arr)
            print(f'auto_restore the duplicates...,loop {_temp_loop_count},\n')
            print(f'duplicates last for:{sum(temp_duplicates_boolean_arr)},\n')

        return temp_nodup_remove_suffix_arr
    else:
        raise ValueError('duplicate_keep can not be None, please reset')


# -----------------------------------------------------------------------------------------
# """level1 Complex DataPrepare Functions"""#################################################
# -----------------------------------------------------------------------------------------

def l1_latent_na_substitute_arr(listarrser, na_regex, case, substitute_value, is_detect_explicit_na):
    """
    detect latent na, like spaces or comma, and substitute the whole element
    :param is_detect_explicit_na: bool, True for substitute explicit na, False for not substitute explicit na
    :param substitute_value: the substitute value to replace the detected na
    :param listarrser: input
    :param na_regex: regex
    :param case: If True, case sensitive.
    :return: bool array
    """
    detect_arr = latent_na_detector_bool_arr(listarrser=listarrser,
                                             na_regex=na_regex,
                                             case=case,
                                             is_detect_explicit_na=is_detect_explicit_na)
    temp_arr = copy.deepcopy(np.array(listarrser))
    # before change, first copy
    temp_arr[detect_arr] = substitute_value
    return temp_arr



import re
import json
import time
import warnings

import nltk
import string
import itertools
import requests
import pandas as pd
from bs4 import BeautifulSoup
from urllib.parse import urljoin
from collections import Counter
from nltk.corpus import stopwords
from nltk.stem.porter import PorterStemmer
from io import BytesIO

import fake_useragent
"""
Author: Romain Boulland, Thomas Bourveau, Matthias Breuer
Date: Jun 24, 2021
Code: Generate bag of words for a time-series of company websites

* Please customize the config file before launching the program *

* Download the following nltk corpus before running the program *
nltk.download('stopwords')
nltk.download('punkt')
"""

"""
modified from origin WaybkJsonToCsvT.py, config.py

Author: Qihang Zhang in 2023/02/03,National University of Singapore,
NUS Business School, Department of Accounting
"""
# %% Macros
# STATUS_CODE = config.status_code
# MIME_TYPE = config.mime_type
# HEADERS = config.headers
# MAX_URL = config.max_url
# MAX_SUB = config.max_sub
# PARSER = config.parser
# RAW = config.raw
# BOW_OPTIONS = config.bow_options
# PATH = config.path


class WayBack:

    def __init__(self, PATH, contact='youremail@domain.com', **kwargs):
        """
        Modified from origin Romain Boulland, Thomas Bourveau, Matthias Breuer's work
        Author: Qihang Zhang in 2023/02/03,National University of Singapore,
        NUS Business School, Department of Accounting
        :param kwargs: useragent/alpha_token/word_len/stop_words/stemmer/STATUS_CODE/MIME_TYPE/
                        HEADERS/MAX_URL/MAX_SUB/PARSER/RAW/BOW_OPTIONS/PATH
                        see:https://github.com/r-boulland/Corporate-Website-Disclosure
        """
        # old-version -> new version, use random fake useragent instead of manuall input>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        """
        useragent = kwargs['useragent'] if 'useragent' in kwargs else 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) ' \
                                                                      'AppleWebKit/537.36 (KHTML, like Gecko) ' \
                                                                      'Chrome/42.0.2311.135 Safari/537.36 Edge/13.10586'
        if not ('useragent' in kwargs):
            warnings.warn('Set your own useragent for safer using, like https://zhuanlan.zhihu.com/p/97973031')
        """
        # old version: 0.1.1.230208_alpha deprecate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

        useragent = kwargs['useragent'] if 'useragent' in kwargs else fake_useragent.UserAgent(
            browsers=["chrome", "edge", "internet explorer", "firefox"]
        ).random
        if not ('useragent' in kwargs):
            print('\n', f'Automatically generate fake useragent:{useragent}', '\n')

        # Some Settings
        alpha_token = kwargs['alpha_token'] if 'alpha_token' in kwargs else True

        word_len = kwargs['word_len'] if 'word_len' in kwargs else(1, 20)

        stop_words = kwargs['stop_words'] if 'stop_words' in kwargs else stopwords.words('english')

        stemmer = kwargs['stemmer'] if 'stemmer' in kwargs else PorterStemmer()

        # Arguments

        self.STATUS_CODE = kwargs['STATUS_CODE'] if 'STATUS_CODE' in kwargs else ['200']

        self.MIME_TYPE = kwargs['MIME_TYPE'] if 'MIME_TYPE' in kwargs else ['text/html']

        self.HEADERS = {'User-Agent': useragent, 'From': contact}

        self.MAX_URL = kwargs['MAX_URL'] if 'MAX_URL' in kwargs else 10

        self.MAX_SUB = kwargs['MAX_SUB'] if 'MAX_SUB' in kwargs else 1

        self.PARSER = kwargs['PARSER'] if 'PARSER' in kwargs else 'lxml'

        self.RAW = kwargs['RAW'] if 'RAW' in kwargs else False

        self.BOW_OPTIONS = kwargs['BOW_OPTIONS'] if 'BOW_OPTIONS' in kwargs else {
            'alpha_token': alpha_token, 'word_len': word_len, 'stop_words': stop_words, 'stemmer': stemmer
        }

        self.PATH = PATH

    @staticmethod
    def wayback_query(host, match_type='exact', collapse_time=10):
        """
        The function wayback_query() interacts with the Wayback Machine API
         to retrieve all matched results for a given input URL.
        The official documentation of the API is here:
        https://github.com/internetarchive/wayback/tree/master/wayback-cdx-server
        """
        # Wayback API query
        q = 'http://web.archive.org/cdx/search/cdx?url={}&matchType={}&collapse=timestamp:{}&output=json' \
            .format(host, match_type, collapse_time)
        r = requests.get(q)
        # Convert json format result into a dataframe
        try:
            df = pd.read_json(r.content)
        except:
            df = pd.read_json(BytesIO(r.content))
        # If no match
        if df.shape[0] == 0:
            print('Wayback does not archive URLs under {}.'.format(host))
            return None
        # If matches returned
        else:
            # First row is column index
            df.columns = df.iloc[0]
            df = df.drop(df.index[0]).reset_index(drop=True)
            print('{} matched URLs found under {}.'.format(df.shape[0], host))
            # Search results cleaning: timestamp\
            # Drop observations with invalid timestamps
            df['datetime'] = pd.to_datetime(df['timestamp'], format='%Y%m%d%H%M%S', errors='coerce')
            if df['datetime'].isnull().any():
                df = df[df['datetime'].notnull()]
                print('{} matches left after dropping invalid timestamps.'.format(df.shape[0]))
            # Search results cleaning: urlkey
            # Ensure uniqueness of urlkey, keep the shortest if not
            if df['urlkey'].nunique() > 1:
                df = df[df['urlkey'] == sorted(df['urlkey'].values, key=len)[0]]
                print('Urlkey not unique, keep the shortest, {} matches left.'.format(df.shape[0]))
            # Complete url for archived websites (as stored on archive.org)
            df['url'] = ['https://web.archive.org/web' + '/' + t + '/' + o for t, o in
                         zip(df['timestamp'].values, df['original'].values)]
            df.sort_values(['urlkey', 'datetime'], inplace=True)
            # Output a txt if specified
            return df

    @staticmethod
    def query_filter(df, freq, date_range=None):
        """
        The function query_filter() refines search results by applying the following filters:
        Filtering search results: date range, status code, MIME type
            - Frequency
            - Date range: Optional
            - Status code: Default
            - MIME type: Default
        """
        if date_range:
            df = df[(df['datetime'] >= pd.to_datetime(date_range[0])) &
                    (df['datetime'] <= pd.to_datetime(date_range[1]))]
        """
        The following filters are deprecated
        A given non-200 URL may be redirected to a valid URL
        Filters applied later based on the final destination
        df = df[df['statuscode'].isin(status_code)]
        df = df[df['mimetype'].apply(lambda x: any([re.search(y, x) for y in mime_type]))]
        """
        # Reset index
        df = df.reset_index(drop=True)
        # Resample the time-series of search results at desired frequency
        df['length'] = pd.to_numeric(df['length'], errors='coerce')
        df_r = df.set_index('datetime').resample(freq)['length'].mean().to_frame().reset_index()
        df_r.columns = ['agg_datetime', 'avg_length']
        df_r['period'] = df_r['agg_datetime'].dt.to_period(freq)
        df['period'] = df['datetime'].dt.to_period(freq)
        df = pd.merge(df, df_r, on=['period'], how='inner')
        df = df.sort_values(['datetime']).drop_duplicates(['period'], keep='last')
        # Print filtering result
        print('{} matches left after filtering.'.format(df.shape[0]))
        # Return filtered dataframe
        return df

    @staticmethod
    def wayback_urlparse(url):
        """
        The function wayback_urlparse() parses the timestamp and original URL for URLs seen during scraping
        """
        # Obtain timestamp and original URL for an archived URL
        if re.search('https://web.archive.org/web/[0-9]{14}/', url):
            timestamp = re.search('[0-9]{14}', url)[0]
            original = re.sub('https://web.archive.org/web/[0-9]{14}/', '', url)
            return timestamp, original
        else:
            return None, None

    @staticmethod
    def text_tokenize(text, alpha_token=False, word_len=None, stop_words=None, stemmer=None):
        """
        The function text_tokenize() preprocesses and tokenizes text into a list of words.
        """
        # Define the punctuation translator
        translator = str.maketrans('', '', string.punctuation)
        # Tokenize sentence into words
        words = nltk.word_tokenize(text)
        # Basic cleaning: lower case, punctuations and nulls removed
        words = [word.lower() for word in words]
        words = [word.translate(translator) for word in words]
        words = [word for word in words if word != '']

        # If needed: length filter, non-alphabetic words, removing stopwords, stemming
        if alpha_token:
            words = [word for word in words if word.isalpha()]
        if word_len:
            words = [word for word in words if word_len[0] <= len(word) <= word_len[1]]
        if stop_words:
            words = [word for word in words if word not in stop_words]
        if stemmer:
            words = [stemmer.stem(word) for word in words]

        # Return the list of processed words
        return words

    def url_scrape(self, url, **kwargs):
        """
        The function url_scrape() scrapes the URL and returns an array of outputs including:
            - HTML text
            - Sub-URLs
            - Error description if occurred
            - Processing time
        :param kwargs: status_code/mime_type/raw/headers/parser
        """
        # change the old input vars be self vars:
        status_code = kwargs['status_code'] if 'status_code' in kwargs else self.STATUS_CODE
        mime_type = kwargs['mime_type'] if 'mime_type' in kwargs else self.MIME_TYPE
        raw = kwargs['raw'] if 'raw' in kwargs else self.RAW
        headers = kwargs['headers'] if 'headers' in kwargs else self.HEADERS
        parser = kwargs['parser'] if 'parser' in kwargs else self.PARSER

        # Dictionary to store parsing results & errors
        url_res = {'URL': url,
                   'cURL': url,
                   'error': None,
                   'subURLs': [],
                   'text': ''}
        if raw:
            url_res['raw'] = ''

        # Request the URL and retrieve status code, mime type, and raw text
        try:
            # Open the sent URL using requests.get()
            curr_request = requests.get(url, headers=headers, allow_redirects=True)
        except Exception as ex_request:
            # Print the exception if requests() fails
            # Mostly due to connection failure or exceeding maximum retries
            url_res['error'] = str(ex_request)
            return url_res

        # The current URL and its timestamp % original
        # May not be the same with the sent URL due to redirection
        curr_url = curr_request.url
        url_res['cURL'] = curr_url
        # Check the status code
        curr_code = curr_request.status_code
        # Check the mime type
        curr_type = curr_request.headers['Content-Type']
        # Open the url and get raw text
        curr_html = curr_request.text
        # Store the raw html text if specified
        if raw:
            url_res['raw'] = curr_html

        # Check status code and MIME type
        try:
            assert str(curr_code) in status_code, 'Invalid HTTP status code: ' + str(curr_code)
            assert any([re.search(x, curr_type) for x in mime_type]), 'Invalid MIME type: ' + str(curr_type)
        except Exception as ex_response:
            url_res['error'] = str(ex_response)
            return url_res

        # Create soup object to remove HTML tags
        curr_soup = BeautifulSoup(curr_html, parser)
        # Kill all script and style elements
        for script in curr_soup(['script', 'style']):
            script.decompose()
        # Kill Wayback bars
        for div_id in ['wm-ipp-base', 'wm-ipp-print', 'donato']:
            for div in curr_soup.find_all('div', id=div_id):
                div.decompose()
        # Body text of the html
        try:
            curr_text = curr_soup.body.get_text()
            url_res['text'] = curr_text
        except Exception as ex_text:
            url_res['error'] = str(ex_text)

        # Get sub-URLs of the URL
        sub_urls = []
        for tag in curr_soup.find_all('a', href=True):
            # Extract the links in the URL
            sub_url = tag['href']
            # Join the sub-URL with the current URL
            # Use the actual current URL instead of the sent URL
            sub_url = urljoin(curr_url, sub_url)
            # Store the sub-URL in a list
            sub_urls.append(sub_url)
        url_res['subURLs'] = sub_urls

        # Print error if there is one
        if url_res['error'] is not None:
            print(url_res['error'])
        # Return a dictionary object for scraping result
        return url_res

    def tree_scrape(self, url, **kwargs):
        """
        The function tree_scrape() scrapes a tree of URLs where the sent URL is the root.
        User needs to specify the max number of URLs (max_url) and/or the max level of sub-URLs (max_sub)
        The function stops when either of the two thresholds is exceeded
        :param kwargs:max_url/max_sub
        """
        max_url = kwargs['max_url'] if 'max_url' in kwargs else self.MAX_URL
        max_sub = kwargs['max_sub'] if 'max_sub' in kwargs else self.MAX_SUB

        # Start time
        start_time = time.time()

        # List to store all scraping results
        tree_res = []

        # Dictionaries to track the tree structure
        # urls_dict: FIFO dictionary sorted by sub-URL level
        # seen_dict: dictionary stored all seen URLs
        urls_dict = {url: 0}
        seen_dict = {url: {'sub': 0, 'info': self.wayback_urlparse(url)}}

        # Navigate the tree
        sub = 0
        while len(urls_dict) > 0 and len(tree_res) < max_url and sub <= max_sub:
            # Get the current URL
            # Pop the one with the lowest sub-level
            curr_url = sorted(urls_dict.items(), key=lambda x: x[1], reverse=False).pop(0)[0]
            # Remove it from the dictionary of all URLs
            urls_dict.pop(curr_url)

            # Parse the current URL
            url_res = self.url_scrape(curr_url)
            # Add sub-level to parsing result
            url_res['sub'] = seen_dict[curr_url]
            # Add to the list of opened URLs
            tree_res.append(url_res)

            # Get sub-URLs of the current URL
            for sub_url in url_res['subURLs']:
                # If sub_url is not in seen, then add it along with its level to urls(dict) and seen(dict)
                sub_timestamp, sub_original = self.wayback_urlparse(sub_url)
                if sub_original is not None and sub_original not in [v['info'][1] for v in seen_dict.values()]:
                    urls_dict[sub_url] = seen_dict[curr_url]['sub'] + 1
                    seen_dict[sub_url] = {'sub': urls_dict[sub_url], 'info': (sub_timestamp, sub_original)}

            # Current level of sub-URL
            sub = min(urls_dict.values(), default=max([v['sub'] for v in seen_dict.values()]))

            # Random waiting time before the next request
            time.sleep(1)

        # End time
        end_time = time.time()

        # Print scraping progress
        print('URL %s finished scraping in %.2f seconds. '
              '%d URLs seen. %d opened. %d returned errors. '
              'Stop at level %d sub-URLs.' %
              (url, end_time - start_time,
               len(seen_dict), len(tree_res), len([d for d in tree_res if d['error'] is not None]),
               sub)
              )

        # Return list of scraping results
        return tree_res

    def wayback_scraper(self, host, freq, date_range=None, **kwargs):
        """
        Main function
        The function WaybackScraper() serves as the main function of the program.
        It generates bag of words for a time-series of company websites.
        :param kwargs: max_url/max_sub/bow_options/path
        """
        max_url = kwargs['max_url'] if 'max_url' in kwargs else self.MAX_URL

        max_sub = kwargs['max_sub'] if 'max_sub' in kwargs else self.MAX_SUB

        bow_options = kwargs['bow_options'] if 'bow_options' in kwargs else self.BOW_OPTIONS

        path = kwargs['path'] if 'path' in kwargs else self.PATH

        # Search all archives for the entered URL
        df = self.wayback_query(host)

        # Break if no result found
        if df is None:
            return

        # Store the API query results
        df.to_csv(path+'cdx['+re.sub(r'[^\w\s\-]', '_', host)+'].csv', index=False)

        # Aggregate query results at desired frequency
        df = self.query_filter(df, freq, date_range)

        # Scrape all URLs and their subs
        url_res_all = {}
        for url in df['url'].values:
            tree_res = self.tree_scrape(url, max_url=max_url, max_sub=max_sub)
            url_res_all[url] = tree_res

        # Store the time-series of scraped results
        with open(path+'res['+re.sub(r'[^\w\s\-]', '_', host)+'].json', 'w') as f:
            json.dump(url_res_all, f)

        # Process scraped html texts
        args = bow_options.values()
        url_bow_all = {
            url: Counter(
                itertools.chain(
                    *[self.text_tokenize(url_res['text'], *args) for url_res in tree_res]
                )
            )
            for url, tree_res in url_res_all.items()
        }

        # Store the time-series of formed BoWs
        with open(path+'bow[' + re.sub(r'[^\w\s\-]', '_', host)+'].json', 'w') as f:
            json.dump(url_bow_all, f)



import pandas as pd
import sas7bdat


def harte_hanks_txt_reader_2019_2021_df(PATH, **kwargs):
    """encoding='ISO-8859-1
    :param kwargs: other arguments should be passed to pd.read_csv()
    '"""
    temp_df = pd.read_csv(filepath_or_buffer=PATH, encoding="ISO-8859-1", low_memory=False, **kwargs)
    return temp_df


def harte_hanks_txt_reader_2017_2018_df(PATH, **kwargs):
    """encoding='UTF-16'
    :param kwargs: other arguments should be passed to pd.read_csv()
    """
    temp_df = pd.read_csv(filepath_or_buffer=PATH, encoding="UTF-16", low_memory=False, **kwargs)
    return temp_df


def harte_hanks_txt_reader_2016_df(PATH, **kwargs):
    """encoding='ISO-8859-1'
    :param kwargs: other arguments should be passed to pd.read_csv()
    """
    temp_df = pd.read_csv(filepath_or_buffer=PATH, encoding="ISO-8859-1", low_memory=False, **kwargs)
    return temp_df


def harte_hanks_txt_reader_1996_2015_df(PATH, **kwargs):
    """encoding='ISO-8859-1'
    :param kwargs: other arguments should be passed to pd.read_csv()
    """
    temp_df = pd.read_csv(
        filepath_or_buffer=PATH, encoding="ISO-8859-1", sep='\t', low_memory=False, **kwargs
    )
    return temp_df


def sas7bdat_reader_df(path, **kwargs):
    """
    read .sas7bdat data by sas7bdat.SAS7BDAT
    :param path: path to read, must have
    :param kwargs: other arguments of sas7bdat.SAS7BDAT(), see https://pyhacker.com/pages/sas7bdat.html
    """
    with sas7bdat.SAS7BDAT(path=path, **kwargs) as result:
        temp_df = result.to_data_frame()

    return temp_df


# -----------------------------------------------------------------------------------------
# """L1 Complex DataPrepare Functions"""#######################################################
# -----------------------------------------------------------------------------------------
def l1_hartehanks_readtxt_df(PATH, year: int, **kwargs):
    """
    :param PATH: read path
    :param year: be int i>=1996 should
    :param kwargs: other arguments should be passed to pd.read_csv()
    :return:
    """
    # dealing kwargs
    if (year >= 1996) and (year <= 2015):
        _temp_read_function = harte_hanks_txt_reader_1996_2015_df
    elif year == 2016:
        _temp_read_function = harte_hanks_txt_reader_2016_df
    elif (year >= 2017) and (year <= 2018):
        _temp_read_function = harte_hanks_txt_reader_2017_2018_df
    elif (year >= 2019) and (year <= 2021):
        _temp_read_function = harte_hanks_txt_reader_2019_2021_df
    else:
        raise ValueError(
            'year {} is not in function range, try again or change the read function'.format(year))

    return _temp_read_function(PATH, **kwargs)



import pandas as pd
import numpy as np


def suffix_remove_arr(listarrser, suffix_regex: str, **kwargs):
    """
    Input the suffix regex to remove the suffix, unlike pd.str.removesuffix
    Anything you input will seem like a group for regex
    :param listarrser: input array-like
    :param suffix_regex: regex string of suffix to remove
    :param kwargs:  pd.series.str.replace's other arguments;regex=True could not be changed
        BASED ON:https://pandas.pydata.org/docs/reference/api/pandas.Series.str.replace.html
    """

    # examine
    if (suffix_regex == '') or (not isinstance(suffix_regex, str)):
        raise ValueError('suffix_regex must be string and should not be ""')

    # core part
    temp_remove_suffix_arr = pd.Series(listarrser).astype(str).str.replace(
        '(' + suffix_regex + '){1}$', repl='', regex=True, **kwargs).values

    return temp_remove_suffix_arr


def extract_prefix_arr(listarrser, prefix_regex):
    """extract prefix by regex"""
    temp_extract_arr = listarrser.astype(str).str.extract(
        f'(^{prefix_regex}).*$', expand=False
    ).values  # choose

    return temp_extract_arr


def restore_leading_zeros_str_arr(listarrser, target_len):
    """
    Restore the lost leading zeros of the int/zeros list/arr/ser
    :param target_len: restore leading zeros to make new string back to target length.
    :param listarrser: INT list/array/ser which may lost leading zeros when saved
    :return: a string type array contains restored leading zeros
    """
    if np.all(pd.Series(listarrser).apply(lambda x: isinstance(x, int))):
        temp_string_ser = pd.Series(listarrser).astype(str)
    elif np.all(pd.Series(listarrser).apply(lambda x: isinstance(x, str))):
        # to judge that if the string of data are int
        temp_string_ser = pd.Series(listarrser).astype(int).astype(str)
    else:
        raise ValueError('Input array like should be int elements, or int strings')

    if np.nanmax(temp_string_ser.apply(len)) > target_len:
        raise ValueError('The max length of elements bigger than target length')

    dealed_string_ser = temp_string_ser.apply(
        lambda x: '0'*(target_len-len(x))+x
    )
    return dealed_string_ser


# -----------------------------------------------------------------------------------------
# """L1 Complex DataPrepare Functions"""#######################################################
# -----------------------------------------------------------------------------------------
def l1_suffix_and_around_remove_arr(listarrser, suffix_regex: str, around_regex: str, **kwargs):
    """
    REMOVE THE suffix, which ALSO remove all possible words or commas around it, like ',';'.';Space
    Anything you input will seem like a group for regex.
    WARNING: for set *, around_regex may work for ZERO TIME and CAUSE unintend results.
    :param listarrser: input array-like
    :param suffix_regex: regex string of suffix to remove
    :param around_regex: around words/comma regex, which are sepS BEFORE AND AFTER the target suffix.
    :param kwargs: pd.series.str.replace's other arguments-> case, flags, etc;regex=True could not be changed
        BASED ON:https://pandas.pydata.org/docs/reference/api/pandas.Series.str.replace.html
    """

    # examine
    if (suffix_regex == '') or (not isinstance(suffix_regex, str)):
        raise ValueError('suffix_regex must be string and should not be ""')

    # core part
    if around_regex == '':
        # temp_remove_suffix_arr = pd.Series(listarrser).astype(str).str.replace(
        #     '(' + suffix_regex + '){1}$', repl='', case=case, regex=True).values
        temp_remove_suffix_arr = suffix_remove_arr(
            listarrser=listarrser,
            suffix_regex=suffix_regex,
            **kwargs
        )
    else:
        temp_remove_suffix_arr = pd.Series(listarrser).astype(str).str.replace(
            f'({around_regex})*'+'(' + suffix_regex + '){1}' + f'({around_regex})*$',
            repl='',
            regex=True,
            **kwargs
        ).values

    return temp_remove_suffix_arr
